from dataclasses import dataclass
from typing import List

@dataclass
class Order:
    price: int
    assetId: str
    amount: int
    type: str

@dataclass
class Metadata:
	name: str
	description: str
	ticker: str
	url: str
	logo: str
	decimals: int


@dataclass
class AssetInfo:
	asset: str
	policy_id: str
	asset_name: str
	fingerprint: str
	quantity: int
	initial_mint_tx_hash: str
	mint_or_burn_count: int
	metadata: Metadata
	onchain_metadata: dict
	asset_id: str

@dataclass
class Pool:
    price:          int
    assetId:        str
    queue:          int
    quantity_ada:   int
    quantity_asset: int
    decimalA:       int
    decimalB:       int

@dataclass
class Asset:
    policy: str
    asset: str
    amount: int

@dataclass
class Input:
    tx_id: str
    index: int

@dataclass
class UTXO:
    address: str
    amount: int
    assets: List[Asset]
    output_index: int

@dataclass
class Transaction:
    hash: str
    fee: int
    ttl: int
    validity_interval_start: int
    network_id: str
    input_count: int
    output_count: int
    mint_count: int
    total_output: int
    metadata: List
    inputs: List[Input]
    outputs: List[UTXO]
    mint: List
    timestamp: int


from binascii import unhexlify
import requests
from pymongo.cursor import Cursor
from .models import Transaction, Asset, Input, UTXO, AssetInfo
from typing import List

LOVELACE = 1000000


def create_transaction_object(oura_transaction: dict) -> Transaction:
    tx_data = oura_transaction['transaction']
    outputs = []
    for output in tx_data['outputs']:
        assets = []
        for asset in output['assets']:
            assets.append(Asset(**asset))
        del output['assets']
        outputs.append(UTXO(**output, assets=assets))

    inputs = [Input(**input) for input in tx_data['inputs']]

    del oura_transaction['transaction']['outputs']
    del oura_transaction['transaction']['inputs']

    transaction = Transaction(**oura_transaction['transaction'], outputs=outputs, inputs=inputs, timestamp=oura_transaction['timestamp'])
    return transaction


def get_decimals(asset_info: AssetInfo):
    if(
        'metadata' in asset_info
        and asset_info['metadata']
        and 'decimals' in asset_info['metadata']
    ):
        return asset_info['metadata']['decimals']
    else:
        return 0


def get_asset_name(asset_info: AssetInfo):
 
    if asset_info['asset_name']:
        name = unhexlify(asset_info['asset_name']).decode()
    else:
        name = 'Unknown'
    return name


def get_transactions(address: str, metadata: str, oura_transactions_collection: Cursor) -> List[Transaction]:
    query = {
        "transaction.outputs.address": address, 
        "transaction.metadata.map_json.msg": metadata
        }
    result = oura_transactions_collection.find(query)
    return [create_transaction_object(tx) for tx in result]


from binascii import unhexlify
from pymongo.cursor import Cursor
from blockfrost import BlockFrostApi
import json
import requests
from .models import AssetInfo, Metadata

LOVELACE = 1000000

def fix_decimals(order, decimal_places):
    
    if decimal_places != 0:
        order['price'] = int((order['price'] * 10**decimal_places) * LOVELACE)
        order['amount'] = order['amount'] / 10**decimal_places

def create_asset(asset_data: dict):
    metadata = None
    if 'metadata' in asset_data and asset_data['metadata'] != None:
        metadata = Metadata(**asset_data['metadata'])
        asset_data['metadata'] = metadata
    del asset_data['_id']
    asset_info = AssetInfo(**asset_data)
    asset_info.metadata = metadata

    return asset_info

def resolve_asset(id, assets_collection: Cursor, blockfrost_api: BlockFrostApi) -> AssetInfo:
    asset_info = [asset for asset in assets_collection.find({"asset_id": id})]
    info = None
    if len(asset_info) == 0:
        blockfrostasset_id = id.replace('.', '') 
        info = blockfrost_api.asset(blockfrostasset_id, return_type='json')
        info['asset_id'] = info['policy_id'] + '.' +  (info['asset_name'] or '')
        assets_collection.insert_one(info)
    
    else:
        info = asset_info[0]
    
    info = create_asset(info)
    return info

def build_whale_message(tx_hash, exchange_name, asset_name, percent_change):
    message = exchange_name + "\n"
    message += f"Whale splash for {asset_name}: {percent_change}%\n"
    message += f'transaction: https://cardanoscan.io/transaction/{tx_hash}\n'

    return message

def send_message(message):
    url = "http://telegram:7000/message"

    payload = json.dumps({
        "text": message
    })

    response = requests.request("POST", url, data=payload)
    assert response.status_code == 200


from typing import Callable, List
from .models import Pool


LOVELACE = 1000000

def get_unique_pools(pools) -> List[Pool]:
    unique_pools = {}
    for pool in pools:
        if pool['assetId'] not in unique_pools:
            unique_pools[pool['assetId']] = pool
        elif (
            pool['assetId'] in unique_pools and 
            (
                unique_pools[pool['assetId']]['volume'] < pool['volume'] or
                (
                    unique_pools[pool['assetId']]['volume'] == pool['volume'] and
                    unique_pools[pool['assetId']]['fee'] > pool['fee']
                )
            )
            
        ):
            unique_pools[pool['assetId']] = pool
    return list(unique_pools.values())

def calculate_impact(asset_quantity, asset_decimals, pool):
    asset_quantity  = float(asset_quantity)
    
    if asset_decimals and asset_decimals != 0:
        asset_quantity = asset_quantity / 10 ** asset_decimals

    quantity_ada    = float(pool['quantity_ada'])
    quantity_asset  = float(pool['quantity_asset'])
    current_price   = quantity_ada / quantity_asset
    new_stored_ada  = quantity_ada * quantity_asset / (quantity_asset + asset_quantity)
    new_price       = new_stored_ada / (quantity_asset + asset_quantity)
    slippage        = new_price / current_price
    percent_change  = (1 - slippage) * 100
    return percent_change

def calculate_lovelace_price(quantityA, quantityB):
    price_ada =  int(quantityA[0]) / int(quantityB[0])
    
    if quantityA[1] != quantityB[1]:
        price_ada = price_ada / 10**quantityA[1] - quantityB[1]
    
    price_lovelace = int(price_ada * LOVELACE)
    
    return price_lovelace

def generate_pools_info(
    pools, 
    get_asset_id:           Callable, 
    get_token_name:         Callable, 
    get_ticker:             Callable,
    get_fee:                Callable,
    get_pool_id:            Callable,
    get_pool_ada_quantity:  Callable,
    get_quantity_asset:     Callable,
    get_decimals_asset:     Callable,
    get_volume:             Callable,
    get_queue:              Callable,
    ) -> list:

    assets = []
    for pool in pools:
        row = {}
        row["assetId"]          = get_asset_id(pool)
        row["tokenName"]        = get_token_name(pool)
        row["ticker"]           = get_ticker(pool)
        row['fee']              = get_fee(pool)
        
        row["ident"]            = get_pool_id(pool)
        row['quantity_ada']     = get_pool_ada_quantity(pool)
        row['quantity_asset']   = get_quantity_asset(pool)
        row['decimal_asset']    = get_decimals_asset(pool)
        row['decimal_ada']      =  6
        
        qA = (row['quantity_asset'], row['decimal_asset'])
        qB = (row['quantity_ada'], row['decimal_ada'])

        row["price"] = calculate_lovelace_price(qA, qB)
        
        row['volume'] = get_volume(pool)
        row['queue'] = get_queue(pool)
        assets.append(row)
    return assets